@@ -447,7 +447,7 @@ try {
 							window.setTimeout(arguments.callee, 10);
 						}
 
-						TSM.restoreAllTabs(loadData);
+						var doneCorrectly = TSM.restoreAllTabs(loadData);
 						var b = TSM.service.browser;
 						if (
 							(
@@ -461,10 +461,10 @@ try {
 							(
 								TSM.service.getPref('browser.tabs.extensions.last_tab_closing') == 3 &&
 								b.mTabs.length == 1 &&
-								b.mCurrentTab.isReallyBlank
+								b.selectedTab.isReallyBlank
 							)
 							)
-							b.removeTabInternal(b.mCurrentTab, { preventUndo : true });
+							b.removeTabInternal(b.selectedTab, { preventUndo : true });
 					}
 					restore();
 				}
@@ -513,7 +513,7 @@ catch(e) {
 				(
 					this.service.browser.mTabs.length == 1 &&
 					this.service.browser.mRemovedTabInfoList.length == 1 &&
-					!this.service.browser.mCurrentTab.isReallyBlank
+					!this.service.browser.selectedTab.isReallyBlank
 				)
 			)
 			) {
@@ -549,10 +549,12 @@ catch(e) {
 		var b = this.service.browsers,
 			j;
 		if (this.service.getPref('browser.tabs.extensions.undo_cache.global')) {
-			for (i = 0; i < b.length; i++)
+			for (i = 0; i < b.length; i++) {
+				if (!b[i].mTabs) continue; // if the browser has no 'mTabs' property, shutdown process wrongly stops on this point.
 				for (j = 0; j < b[i].mTabs.length; j++)
 					if (!b[i].mTabs[j].isReallyBlank)
 						this.updateGlobalUndoCache(b[i].getTabInfo(b[i].mTabs[j]), true);
+			}
 		}
 
 		if (this.service.isBrowserWindow) {
@@ -588,7 +590,7 @@ catch(e) {
 		}
 	},
  
-	onTabReordered : function(aTabId, aIndex) 
+	onTabMoved : function(aTabId, aIndex) 
 	{
 		// move the entry for the tab
 		if (this.shouldBackupTabs)
@@ -598,8 +600,8 @@ catch(e) {
 	// edit the global "undo close tab" cache when the cache in this window has been modified 
 	onXULTabbrowserUndoCacheModified : function(aEvent, aInfo)
 	{
-		if (aEvent.targetTabInfo)
-			TabbrowserSessionManager.updateGlobalUndoCache(aEvent.targetTabInfo, aEvent.type == 'XULTabbrowserUndoCacheAdded')
+		if (aEvent.cachedTabInfo || aEvent.removedCachedTabInfo)
+			TabbrowserSessionManager.updateGlobalUndoCache(aEvent.cachedTabInfo || aEvent.removedCachedTabInfo, aEvent.type == 'XULTabbrowserUndoCacheAdded');
 	},
 	
 	updateGlobalUndoCache : function(aInfo, aAdding) 
@@ -641,7 +643,7 @@ catch(e) {
 
 		gTSGlobalUndoRemoveTabCacheDatabase.reset();
 		for (var i = 0; i < gTSGlobalUndoRemoveTabCacheDatabase.length; i++) {
-			gTSGlobalUndoRemoveTabCache[i] = this.loadTabInfoFrom(gTSGlobalUndoRemoveTabCacheDatabase.item(i), gTSGlobalUndoRemoveTabCacheDatabase);
+			gTSGlobalUndoRemoveTabCache[i] = this.loadTabInfoFrom(gTSGlobalUndoRemoveTabCacheDatabase.item(i), gTSGlobalUndoRemoveTabCacheDatabase, {});
 		}
 	},
    
@@ -660,18 +662,6 @@ catch(e) {
 			'selected',           aInfo.selected,
 			'HistoryIndex',       aInfo.SHIndex,
 
-			'locked',             aInfo.locked,
-			'referrerBlocked',    aInfo.referrerBlocked,
-			'autoreloadInterval', String(aInfo.autoreloadInterval),
-			'autoreloadPostType', aInfo.autoreloadPostType,
-			'autoreloadPostData', this.service.escape(aInfo.autoreloadPostData || ''),
-
-			'allowPlugins',       aInfo.allowPlugins,
-			'allowJavascript',    aInfo.allowJavascript,
-			'allowMetaRedirects', aInfo.allowMetaRedirects,
-			'allowSubframes',     aInfo.allowSubframes,
-			'allowImages',        aInfo.allowImages,
-
 			'fixedLabel',         aInfo.fixedLabel,
 			'textZoom',           aInfo.textZoom,
 
@@ -822,7 +812,7 @@ catch(e) {
 }
 	},
   
-	loadTabInfoFrom : function(aRes, aDatabase) 
+	loadTabInfoFrom : function(aRes, aDatabase, aError) 
 	{
 		var info = {
 			id                 : aDatabase.getData(aRes, 'tabId'),
@@ -830,24 +820,12 @@ catch(e) {
 			label              : aDatabase.getData(aRes, 'label'),
 			uri                : aDatabase.getData(aRes, 'uri'),
 			selected           : (aDatabase.getData(aRes, 'selected') == 'true'),
-			SHEntries          : this.getSHEntriesFrom(aRes, aDatabase),
+			SHEntries          : this.getSHEntriesFrom(aRes, aDatabase, aError),
 			SHIndex            : Number(aDatabase.getData(aRes, 'HistoryIndex')),
 
 			parentTab          : (aDatabase.getData(aRes, 'parentTab') || null),
 			childTabs          : [],
 
-			locked             : (aDatabase.getData(aRes, 'locked') == 'true'),
-			referrerBlocked    : (aDatabase.getData(aRes, 'referrerBlocked') == 'true'),
-			autoreloadInterval : Number(aDatabase.getData(aRes, 'autoreloadInterval')),
-			autoreloadPostType : aDatabase.getData(aRes, 'autoreloadPostType'),
-			autoreloadPostData : this.service.unescape(aDatabase.getData(aRes, 'autoreloadPostData') || ''),
-
-			allowPlugins       : (aDatabase.getData(aRes, 'allowPlugins') == 'true'),
-			allowJavascript    : (aDatabase.getData(aRes, 'allowJavascript') == 'true'),
-			allowMetaRedirects : (aDatabase.getData(aRes, 'allowMetaRedirects') == 'true'),
-			allowSubframes     : (aDatabase.getData(aRes, 'allowSubframes') == 'true'),
-			allowImages        : (aDatabase.getData(aRes, 'allowImages') == 'true'),
-
 			fixedLabel         : aDatabase.getData(aRes, 'fixedLabel'),
 			textZoom           : Number(aDatabase.getData(aRes, 'textZoom')),
 
@@ -869,7 +847,7 @@ catch(e) {
 	loadTabInfoExtra : [],
 	loadTabInfoIdTable : [],
 	
-	getSHEntriesFrom : function(aParentResource, aDatabase) 
+	getSHEntriesFrom : function(aParentResource, aDatabase, aError) 
 	{
 try {
 		var entries = [],
@@ -892,9 +870,14 @@ try {
 				loadType        : parseInt(this.getLiteral(aDatabase, child, this.kRDF_loadType)),
 				x               : parseInt(this.getLiteral(aDatabase, child, this.kRDF_x)),
 				y               : parseInt(this.getLiteral(aDatabase, child, this.kRDF_y)),
-				children        : this.getSHEntriesFrom(child, aDatabase)
+				children        : this.getSHEntriesFrom(child, aDatabase, aError)
 			};
 
+			if (!entries[entries.length-1].uri) {
+				entries[entries.length-1].uri = 'about:blank';
+				if (aError) aError.value = true;
+			}
+
 			cacheKey    = parseInt(this.getLiteral(aDatabase, child, this.kRDF_cacheKey));
 			postContent = escape(this.service.WalletService ? this.service.WalletService.WALLET_Encrypt(this.getLiteral(aDatabase, child, this.kRDF_postContent)) : this.getLiteral(aDatabase, child, this.kRDF_postContent) );
 			try {
@@ -945,13 +928,13 @@ try {
 		aDatabase.clearData();
 
 		var t = this.service.browser.mTabs;
-		if (t.length < 1) return;
+		if (!t || t.length < 1) return;
 
 		for (i = 0; i < t.length; i++)
 			this.saveTabAt(i, aDatabase);
 }
 catch(e) {
-	if (this.service.debug) alert('in "TabbrowserService.saveAllTabs()"\n\n'+e);
+	if (this.service.debug) alert('in "TabbrowserSessionManager.saveAllTabs()"\n\n'+e);
 }
 	},
 	
@@ -985,12 +968,13 @@ catch(e) {
 	},
   
 	// SẴ^u̓eČ 
-	restoreAllTabs : function(aDatabase)
+	restoreAllTabs : function(aDatabase, aShouldNotAlert)
 	{
+		var doneCorrectly = true;
 try {
-		if (this.service.browsers.length < 1) return;
+		if (this.service.browsers.length < 1) return doneCorrectly;
 
-		if (!aDatabase || !aDatabase.length) return;
+		if (!aDatabase || !aDatabase.length) return doneCorrectly;
 
 		var i,
 			res,
@@ -1025,25 +1009,38 @@ try {
 		}
 		else {
 			var b = this.service.browser,
-				t = b.mCurrentTab;
+				t = b.selectedTab;
 
 			b.removeAllTabsButInternal(t); // close other tabs
 
 //			window.setTimeout(function(aManager, aDatabase, aTab) { aManager.restoreAllTabsFrom(aDatabase, aTab); }, 0, this, aDatabase, t);
-			this.restoreAllTabsFrom(aDatabase, t);
+			if (!this.restoreAllTabsFrom(aDatabase, t)) doneCorrectly = false;
 		}
 }
 catch(e) {
+		doneCorrectly = false;
 	if (this.service.debug) alert('in "TabbrowserSessionManager.restoreAllTabs()"\n\n'+e);
 }
+
+		// broken data
+		if (
+			!aShouldNotAlert &&
+			!doneCorrectly &&
+			this.confirmToClearData() == 0
+			) {
+			this.clearDatabase(aDatabase);
+		}
+
+		return doneCorrectly;
 	},
 	
 	// EBhEƂ̃qXgAꂼ̃EBhEŊJ 
 	restoreAllTabsFromRoot : function(aRoot, aEntries)
 	{
+		var doneCorrectly = true;
 try {
 		var b = TabbrowserService.browser,
-			t = b.mCurrentTab,
+			t = b.selectedTab,
 			i, j;
 
 		var doneFirst = false;
@@ -1053,7 +1050,8 @@ try {
 				doneFirst = true;
 			}
 
-			TabbrowserSessionManager.restoreAllTabsFrom(aEntries[i], i == 0 ? b.mCurrentTab : null );
+			if (!TabbrowserSessionManager.restoreAllTabsFrom(aEntries[i], i == 0 ? b.selectedTab : null ))
+				doneCorrectly = false;
 
 			for (j = aEntries[i].length-1; j > -1; j--)
 				TabbrowserSessionManager.removeSHEntriesFrom(aEntries[i].item(j), aEntries[i]);
@@ -1062,15 +1060,26 @@ try {
 		}
 }
 catch(e) {
+		doneCorrectly = false;
 	if (TabbrowserService.debug) alert('in "TabbrowserSessionManager.restoreAllTabsFromRoot()"\n\n'+e);
 }
+
+		// broken data
+		if (
+			!doneCorrectly &&
+			this.confirmToClearData() == 0
+			) {
+			this.clearDatabase(aDatabase);
+		}
 	},
  
 	// qXg^uČ 
 	restoreAllTabsFrom : function(aDatabase, aRemoveTab)
 	{
+		var doneCorrectly = true;
+		var errorOut = { value : false };
 try {
-		if (!aDatabase || !aDatabase.length) return;
+		if (!aDatabase || !aDatabase.length) return doneCorrectly;
 
 		var b = this.service.browser,
 			lastTab,
@@ -1082,7 +1091,7 @@ try {
 		{
 try {
 			res = aDatabase.item(i);
-			info = this.loadTabInfoFrom(res, aDatabase);
+			info = this.loadTabInfoFrom(res, aDatabase, errorOut);
 			info.index = i;
 
 			lastTab = b.addTabWithTabInfo(info);
@@ -1122,13 +1131,19 @@ try {
 if (this.service.debug) dump('<=['+i+'] '+info.uri.substr(0, 25)+' RESTORED.\n');
 }
 catch(e) {
+		doneCorrectly = false;
 	if (this.service.debug) alert('in "TabbrowserSessionManager.restoreAllTabsFrom()"\n'+res.Value+'\n\n'+e);
 }
 		}
 }
 catch(ex) {
+		doneCorrectly = false;
 	if (this.service.debug) alert('in "TabbrowserSessionManager.restoreAllTabsFrom()"\n\n'+ex);
 }
+
+		if (errorOut.value) doneCorrectly = false;
+
+		return doneCorrectly;
 	},
    
 	// ^uZbVz_[̏ 
@@ -1165,7 +1180,16 @@ catch(ex) {
 			b.forwardBrowserGroup = [];
 		}
 
-		this.restoreAllTabs(this.storedTabSets[aID]);
+		var res;
+		if (
+			!this.restoreAllTabs(this.storedTabSets[aID], true) &&
+			this.storedTabSetsRoot.getData(this.storedTabSets[aID].containerNode, 'brokenButNeverAlert') != 'true' &&
+			(res = this.confirmToClearData(true)) == 0
+			) {
+			this.removeStoredTabSet(aID);
+		}
+		if (res == 2)
+			this.storedTabSetsRoot.setData(this.storedTabSets[aID].containerNode, 'brokenButNeverAlert', 'true');
 	},
  
  	renameStoredTabSet : function(aID, aEvent) 
@@ -1221,6 +1245,13 @@ catch(ex) {
 		var TSM = TabbrowserSessionManager;
 		TSM.saveAllTabs(TSM.storedTabSets[aID]);
 
+		// failsafe
+		if (!TSM.storedTabSetsRoot.length) {
+			TSM.storedTabSetsRoot.removeResource(TSM.storedTabSetsRoot.containerNode);
+			TSM.storedTabSetsRoot.makeContainer();
+			TSM.storedTabSetsRoot.reset();
+		}
+
 		TSM.storedTabSetsRoot.setData(TSM.storedTabSets[aID].containerNode, 'Label', aLabel);
 
 		TSM.service.ObserverService.notifyObservers(window, 'tabextensions:storedTabSetsModified', null);
@@ -1388,7 +1419,7 @@ catch(ex) {
 
 		if (!info) return;
 
-		var t      = this.service.browser.mCurrentTab;
+		var t      = this.service.browser.selectedTab;
 		var newTab = this.service.browser.addTabWithTabInfo(info);
 
 		if (newTab && 'restored' in info && info.restored)
@@ -1467,8 +1498,47 @@ catch(ex) {
 			full.removeAttribute('hidden');
 		else
 			full.setAttribute('hidden', true);
-	}
+	},
   
+	confirmToClearData : function(aCanBeIgnored)
+	{
+		return this.service.PromptService.confirmEx(
+				window,
+				this.strbundle.GetStringFromName('confirmToClearBrokenData_title'),
+				this.strbundle.GetStringFromName('confirmToClearBrokenData_text'),
+				(
+					(this.service.PromptService.BUTTON_TITLE_IS_STRING * this.service.PromptService.BUTTON_POS_0) +
+					(this.service.PromptService.BUTTON_TITLE_IS_STRING * this.service.PromptService.BUTTON_POS_1) +
+					(aCanBeIgnored ? (this.service.PromptService.BUTTON_TITLE_IS_STRING * this.service.PromptService.BUTTON_POS_2) : 0 )
+				),
+				this.strbundle.GetStringFromName('confirmToClearBrokenData_clear'),
+				this.strbundle.GetStringFromName('confirmToClearBrokenData_cancel'),
+				(aCanBeIgnored ? this.strbundle.GetStringFromName('confirmToClearBrokenData_ignore') : null ),
+				null,
+				{}
+			);
+	},
+	clearDatabase : function(aDatabase)
+	{
+		var i, j;
+		if (aDatabase == this.tabsBackupRoot ||
+			aDatabase == this.tabHistoryQuitRoot) {
+			for (i = aDatabase.length-1; i > -1; i--)
+			{
+				for (j = aDatabase.item(i).length-1; j > -1; j--)
+					this.removeSHEntriesFrom(aDatabase.item(i).item(j), aDatabase.item(i));
+				aDatabase.item(i).clearData();
+				aDatabase.removeData(aDatabase.item(i).containerNode);
+			}
+		}
+		else {
+			for (i = aDatabase.length-1; i > -1; i--)
+				this.removeSHEntriesFrom(aDatabase.item(i), aDatabase);
+			aDatabase.clearData();
+		}
+		aDatabase.removeResource(aDatabase.containerNode);
+	}
+ 
 }; 
  
 // pref listeners 
