@@ -13,9 +13,10 @@ pref('browser.tabs.opentabfor.linkclick', true);
 pref('browser.tabs.opentabfor.bookmarks', true);
 pref('browser.tabs.opentabfor.history', false);
 pref('browser.tabs.opentabfor.windowopen', true);
-pref('browser.tabs.opentabfor.native_apps', true);
 
-pref('browser.tabs.extensions.native_apps_any', true);
+pref("browser.tabs.extensions.platform_native.behavior", 2);
+
+
 
 
 // Startup
@@ -35,7 +36,7 @@ pref('browser.tabs.extensions.loadInBackgroundMiddleClick', true);
 pref('browser.tabs.extensions.loadInBackgroundBookmarks', true);
 pref('browser.tabs.extensions.loadInBackgroundHistory', true);
 pref('browser.tabs.extensions.loadInBackgroundJS', true);
-pref('browser.tabs.extensions.loadInBackgroundNative', true);
+pref('browser.tabs.extensions.loadInBackgroundPlatformNative', true);
 pref('browser.tabs.extensions.direction_of_focusing', 0);
 
 
@@ -46,3 +47,6 @@ pref('browser.tabs.opentabfor.urlbar_always', false);
 pref('browser.tabs.opentabfor.anybookmark', true);
 pref('browser.tabs.opentabfor.anyhistory', true);
 
+
+// Misc
+pref("browser.tabs.extensions.ignore_target", true);
